/* ndk - [ ndk.c ]
 *
 * Contains the main source file for the ndk
 * operating system.  This is where the system
 * comes alive!
 *
 * Please see:
 *   /doc/*
 *   /src/*
 *
 * (c)2002 dcipher / neuraldk
 *           www.neuraldk.org
 */

#include "multiboot.h"
#include "pic.h"
#include "idt.h"
#include "gdt.h"
#include "task.h"
#include "vm86.h"
#include "rZero.h"
#include "pager.h"
#include "pmode.h"
#include "memory.h"
#include "descgate.h"

// search for VIDINC to make tasks increment video memory again

/* The following are for the multitasking test
 * I can't use local variables, because they're
 * assigned to the stack in C, and I haven't
 * defined separate stacks for each tasks yet ;)
 */
char *videoOne = (char *)0x000B8000 + 0xA0;
char *videoTwo = (char *)0x000B8002 + 0xA0;
char *videoThree = (char *)0x000B8004 + (0xA0 * 11);
char *videoFour  = (char *)0x000b8004 + (0xA0 * 12);

#define TASK_STACK_SIZE 4096
#define MAX_TASKS	5

Task task[MAX_TASKS];
TSS  tss[MAX_TASKS];
char *p0_stack[MAX_TASKS][TASK_STACK_SIZE];
char *p3_stack[MAX_TASKS][TASK_STACK_SIZE];

/*
char *stackOne[TASK_STACK_SIZE];
char *stackTwo[TASK_STACK_SIZE];
char *stackThree[TASK_STACK_SIZE];
Task taskOne, taskTwo, taskMain, taskGateTask;
TSS  tssOne,  tssTwo,  tssMain, tssGateTask;
*/

void jobOne(void);
void jobTwo(void);
void jobGateTask(void);

void temp(void);

int bssCheck;

char *regPrint = "->   Reg: 0x%x   .\n";

void ndkStart(unsigned long magic, unsigned long addr) {
  /* div by zero test! */
  void (*rZeroFunction)(void);
  int x = 3, y = 0;
  short *rmIDT = (short *)0x00001000;
  short c;
  long cs, eip;
  long *pages = (long *)0xffc00000;

  int i;

  long linearAddr;
  long physAddr;

  //stop: goto stop;

  /* for the time being, I'll keep this here */
  //mb_main(magic, addr);

  cls();
  multibootInit(magic, addr);
  pagerInit();
  idtInit();
  /* setup the PIC */
  picInit();
  rZeroInit();
  timerInit();
  vm86Init();
  memoryInit();


  /*
  printf("Allocating a physical page!\n");
  physAddr = (long)memoryAllocUnmappedPage();
  printf("physical address: 0x%8x\n", physAddr);
  physAddr = (long)memoryAllocUnmappedPage();
  printf("physical address: 0x%8x\n", physAddr);
  physAddr = (long)memoryAllocUnmappedPage();
  printf("physical address: 0x%8x\n", physAddr);
  physAddr = (long)memoryAllocUnmappedPage();
  printf("physical address: 0x%8x\n", physAddr);
  memoryStackSpecs(&appStack);
  */

  // allocate 5 pages
  for(i = 0; i < 5; i++) {
    linearAddr = (long)memoryAllocPage();
    //linearAddr = 0x80000000;
    printf("linear addr: 0x%8x\n", linearAddr);

    physAddr = (long)pagerLinToPhys((void *)(linearAddr + 4));
    printf("from pagerLinToPhys: 0x%8x\n", physAddr);
  }

  printf("Writting to page!\n");
  *(long *)linearAddr = 0xffffff;

  //printf("Allocating another page!\n");
  //linearAddr = (long)memoryAllocPage();
  //printf("linear addr: 0x%8x\n", linearAddr);

  stackDisplay(&appStack.used);

  printf("Deallocating first page!\n");
  memoryFreePage((void *)0x80000000);
  //memoryFreePage(linearAddr);

  stackDisplay(&appStack.used);

  // will cause a page fault :)
  //printf("Trying to write to page again!\n");
  //*(long *)linearAddr = 0xffffffff;

  printf("Allocating another page!\n");
  linearAddr = (long)memoryAllocPage();
  printf("linear addr:   0x%8x\n", linearAddr);
  physAddr = (long)pagerLinToPhys((void *)linearAddr);
  printf("physical addr: 0x%8x\n", physAddr);

  //memoryFreePage(0x80000000);

  /*
  for(i = 0; i < 4095; i++)
    printf("0x%x ", pages[i]);
  */

  //linearAddr = memoryAllocatePages(1);
  //printf("Newly allocate space: 0x%8x\n", linearAddr);
  //printf("bssCheck: 0x%x\n", &bssCheck);

  // enable alignment flag for kernel
  asm(
    "pushf\n"
    "pop %eax\n"
    "or $262144, %eax\n"
//  IOPL 3
//    "or $0x3000, %eax\n"
    "push %eax\n"
    "popf\n"
  );

  //printf("main: %x\n", rZeroImports[0].symbolLocation);
  //printf("splashScreen: %x\n", rZeroImports[4].symbolLocation);
  //printf("rZero...\n");
  rZeroFunction = (void (*)(void))rZeroImports[4].symbolLocation;
  //x = x / y;
  /*
  c = 0;
  for(i = 0; i < 0x11; i++) {
    eip  = rmIDT[i*2];
    cs = rmIDT[i*2+1];
    printf("%x, %x:%x\n", i, (long)(cs & 0x0000ffff), (long)(eip & 0x0000ffff));
  }
  */
  //printf("0x%x\n", vm86IntHandler);
  //printf("0x%x\n", &iretInstruction);
  //vm86Test();
  //rZeroFunction();
  //for(;;) {
  //printf("%d\n", timerGetTicks());
  //}

  /* Exception handler tests... :)
  //x = x / y;

  // test
  createDescriptor(&GDT, 4, 0x0, 0xffffff, D_NOT_PRESENT);
  asm(
    "mov $0x11111111, %eax\n"
    "mov $0x22222222, %ebx\n"
    "mov $0x33333333, %ecx\n"
    "mov $0x44444444, %edx\n"
    "mov $0x12345678, %esi\n"
    "mov $0x87654321, %edi\n"
    "ljmp $0x28, $0x87654321");
  */

  //taskFillTSS(&tss[0], 0, 0);
  tss[0].ldtr = 0;
  tss[0].fs = tss[0].gs = 0;
  tss[0].ds = tss[0].es = SEL_P0_DATA;
  tss[0].ss = SEL_P0_STACK;
  tss[0].cs = SEL_P0_CODE;
  tss[0].eflags = 0x202L | 1 << 18;  // IOPL0, STI, Alignment Check
  tss[0].esp = 0;
  tss[0].eip = 0;
  tss[0].cr3 = getCR3();
  taskCreate(&task[0], &tss[0]);
  //printf("Task Main: 0x%x\n", task[0].selector);

  //printf("CR3: 0x%x\n", getCR3());

  //printf("Loading TR\n");
  loadTaskReg(task[0].selector);

  /*
  taskFillTSS(&tss[3], (long)temp,
              (long)p0_stack[3] + TASK_STACK_SIZE,
	      (long)p3_stack[3] + TASK_STACK_SIZE);
  tss->eflags = 0x202L | 1 << 18;  // PL0, STI, Alignment Check
  tss[3].ds = tss[3].es = SEL_P0_DATA | 3;
  tss[3].ss = SEL_P0_STACK | 3;
  tss[3].cs = SEL_P0_CODE | 3;

  //taskCreate(&taskGateTask, &tssGateTask);
  taskCreateSelector(&task[3]);
  createGate(&IDT, 128, 0, task[3].selector/8, D_TASK);
  */

  printf("Creating task #1\n");
  taskFillTSS(&tss[1], (long)jobOne, (long)p0_stack[1] + TASK_STACK_SIZE,
                                     (long)p3_stack[1] + TASK_STACK_SIZE);
  taskCreate(&task[1], &tss[1]);
  //printf("Task 1: 0x%x\n", task[1].selector);


  printf("Creating task #2\n");
  taskFillTSS(&tss[2], (long)jobTwo, (long)p0_stack[2] + TASK_STACK_SIZE,
                                     (long)p3_stack[2] + TASK_STACK_SIZE);
  taskCreate(&task[2], &tss[2]);
  //printf("Task 2: 0x%x\n", task[2].selector);

  // make a call gate
  //printf("Creating task gate\n");
  /*
  taskFillTSS(&tss[3], (long)temp, (long)p0_stack[3] + TASK_STACK_SIZE,
                                     (long)p3_stack[3] + TASK_STACK_SIZE);
  //taskCreate(&task[3], &tss[3]);
  taskCreateTaskGate(&task[3], &tss[3]);
  */
  rZeroFunction();

  printf("Video loop\n");
  loopFour:
    //VIDINC (*videoFour)++;
    //asm("int $0x0\n");
    goto loopFour;


  asm("stop: \n hlt \n jmp stop\n");
  return;
}

/* multitasking test!!! */
void jobOne(void) {
  // Should NOT execute, because of IOPL!
  //asm("cli");
  //printf("Task Gate Link: 0x%x, EFlags: 0x%x", tss[3].link, tss[3].eflags);
  //if(tss[3].eflags & 16384)
  //  printf("NT\n");
  //else printf("\n");

  //asm("int $128\n");
  loopOne:
    //VIDINC (*videoOne)++;
    // for overflow flag... just a test
    /*
    asm("movl $0xffffffff, %%eax\n"
        "addl $0x1, %%eax\n"
        :
        :
        : "eax");
    */
    goto loopOne;
}

void jobTwo(void) {
  //printf("Task Gate Link: 0x%x, EFlags: 0x%x", tss[3].link, tss[3].eflags);
  /*
    asm(
    //"pushl %0\n"
    "pushl %1\n"
    "pushl %0\n"
    "call printf\n"
    "addl $8, %%esp\n"
    :
    : "g" (regPrint), "g" (tss[3].eflags)
  );
  if(tss[3].eflags & 16384)
    printf("NT\n");
  else printf("\n");

  */
  //printf("task 1 eflags: 0x%x .\n", tss[1].eflags);
  loopTwo:
    // VIDINC (*videoTwo)++;
    //asm("int $128\n");
    goto loopTwo;
}

/*
asm(
  ".global temp\n"
  "temp:\n"
  "jmp temp\n"
  */
  /*
  "push %eax\n"
  "push %ds\n"
  "pusha\n"
  "mov $0x08, %ax\n"
  "mov %ax, %ds\n"
  "call jobGateTask\n"
  "popa\n"
  "pop %ds\n"
  "pop %ax\n"
  */
  /*
  "iretl\n"
  // following line needed because it's a task gate!!!
  "jmp temp\n");
  */

/* This is a task gate... somewhere here this gate must be
 * added to the list of on-going tasks...
 */
 /*
void jobGateTask(void) {
  //int i;
  //for(i = 0; i < 1000; i++)
  //printf("->Task Gate Link: 0x%x, EFlags: 0x%x\n", tss[3].link, tss[3].eflags);

  asm(
    //"pushl %0\n"
    "pushf\n"
    "pushl %0\n"
    "call printf\n"
    "addl $8, %%esp\n"
    :
    : "g" (regPrint)
  );
  loopThree:
    (*videoThree)++;
    goto loopThree;
}*/
