/* ndk - [ string.c ]
 *
 * Contains functions similar to the Ansi C
 * implementations in memory.h and string.h
 * Many of these have been borrowed from
 * the NetBSD C Library.  Not only is development
 * quicker when you can rely on these open source
 * projects, it's also incredibly stable (how many
 * thousands of users do you think have "beta
 * tested" those NetBSD routines?)
 *
 * Of course, some of these routines may have to
 * be replaced if I go with a non-GPL license. I
 * haven't really decided upon licensing yet...
 *
 * Please see:
 *   http://www.ajk.tele.fi/libc/code.html [ NetBSD C Library Site ]
 *
 * (c)2002 dcipher / neuraldk
 *           www.neuraldk.org
 */

#include "string.h"

/*
 * sizeof(word) MUST BE A POWER OF TWO
 * SO THAT wmask BELOW IS ALL ONES
 */
typedef long word;              /* "word" used for optimal copy speed */

#define wsize   sizeof(word)
#define wmask   (wsize - 1)


int strncmp(char *s1, char *s2, long n) {

  if(n == 0)
    return 0;

  do {
    if (*s1 != *s2++)
      return (*(unsigned char *)s1 -
              *(unsigned char *)--s2);

    if (*s1++ == 0) break;
  } while (--n != 0);

  return 0;
}


int strcmp(char *s1, char *s2) {
  while (*s1 == *s2++)
    if (*s1++ == 0)
      return 0;

  return (*(char *)s1 - *(char *)--s2);
}


void *memcpy(void *dst0, void *src0, long length) {
//        void *dst0;
//        const void *src0;
//        register size_t length;

  char *dst = dst0;
  const char *src = src0;
  long t;

  if (length == 0 || dst == src)          /* nothing to do */
    goto done;

        /*
         * Macros: loop-t-times; and loop-t-times, t>0
         */
#define TLOOP(s) if (t) TLOOP1(s)
#define TLOOP1(s) do { s; } while (--t)

  if ((unsigned long)dst < (unsigned long)src) {
    /*
     * Copy forward.
     */
    t = (long)src;  /* only need low bits */
    if ((t | (long)dst) & wmask) {
      /*
       * Try to align operands.  This cannot be done
       * unless the low bits match.
       */
      if ((t ^ (long)dst) & wmask || length < wsize)
        t = length;
      else
        t = wsize - (t & wmask);
      length -= t;
      TLOOP1(*dst++ = *src++);
    }
    /*
     * Copy whole words, then mop up any trailing bytes.
     */
    t = length / wsize;
    TLOOP(*(word *)dst = *(word *)src; src += wsize; dst += wsize);
    t = length & wmask;
    TLOOP(*dst++ = *src++);
  } else {
    /*
     * Copy backwards.  Otherwise essentially the same.
     * Alignment works as before, except that it takes
     * (t&wmask) bytes to align, not wsize-(t&wmask).
     */
    src += length;
    dst += length;
    t = (long)src;
    if ((t | (long)dst) & wmask) {
      if ((t ^ (long)dst) & wmask || length <= wsize)
        t = length;
      else
        t &= wmask;
      length -= t;
      TLOOP1(*--dst = *--src);
    }
    t = length / wsize;
    TLOOP(src -= wsize; dst -= wsize; *(word *)dst = *(word *)src);
    t = length & wmask;
    TLOOP(*--dst = *--src);
  }
done:
  return (dst0);
}


void *memset(void *dst, int c, int n) {
  if (n != 0) {
    char *d = dst;

    do {
      *d++ = c;
    } while (--n != 0);
  }
  return (dst);
}
