/* ndk - [ task.c ]
 *
 * This file encompasses the ndk multitasking
 * code.  This code is currently being revamped
 * and will, eventually, be ported to an rZero
 * module to boot-time interchangeability.
 *
 * Please see:
 *   /doc/commonTaskStructure.kwd
 *
 * (c)2002 dcipher / neuraldk
 *           www.neuraldk.org
 */

#include "task.h"

Task *currentTask = NULL;
Task *firstTask = NULL;

void taskFillTSS(TSS *tss, long eip, long p0_esp, long p3_esp) {
  tss->ldtr = 0;
  tss->fs = tss->gs = SEL_P3_DATA | 3;
  tss->ds = tss->es = SEL_P3_DATA | 3;
  tss->ss = SEL_P3_STACK | 3;
  tss->cs = SEL_P3_CODE | 3;
  /*
  tss->fs = tss->gs = SEL_P0_DATA;
  tss->ds = tss->es = SEL_P0_DATA;
  tss->ss = SEL_P0_STACK;
  tss->cs = SEL_P0_CODE;
  */
  tss->eflags = 0x00000202L | 1 << 18;  // IOPL=0, STI, Alignment Check
  tss->esp = /*p0_esp;*/p3_esp;
  tss->eip = eip;
  tss->cr3 = getCR3();

  tss->esp0 = p0_esp;
  tss->ss0 = SEL_P0_STACK;

  tss->link = 0;
}

long taskFindEmptySelector(void) {
  // find first empty GDT entry (remember, GDT entry 0
  // is zero, but we don't want to use it, hense the 2)
  return arrayFindNthEmptyElement(&GDT, 2);
}

long taskCreateSelector(Task *task) {
  int selectorIndex = taskFindEmptySelector();

  task->selector = selectorIndex * 8;

  createDescriptor(&GDT, selectorIndex, task->tss, sizeof(TSS), D_TSS);
}

void taskCreate(Task *task, TSS *tss) {
  Task *llTask;

  task->tss = (long)tss;
  taskCreateSelector(task);

  if(firstTask == NULL) {
    firstTask = (Task *)task;
    currentTask = (Task *)task;
    return;
  }

  llTask = firstTask;
  while(llTask->next != NULL) llTask = (Task *)llTask->next;
  llTask->next = (long)task;

  return;
}

void taskCreateTaskGate(Task *task, TSS *tss) {
  Task *llTask;
  int selectorIndex;

  task->tss = (long)tss;
  taskCreateSelector(task);	// create selector in GDT

  //selectorIndex = 128;
  //task->selector = selectorIndex * 8;
  createGate(&IDT, 128, 0x0, task->selector | 3, D_TASK + D_DPL3);

  /* do not add take gate to linked list!!!
  if(firstTask == NULL) {
    firstTask = (Task *)task;
    currentTask = (Task *)task;
    return;
  }

  llTask = firstTask;
  while(llTask->next != NULL) llTask = (Task *)llTask->next;
  llTask->next = (long)task;
  */

  return;
}

int taskNotBusy(Task *task) {
  char *desc;
  int temp;

  /* check to see if current task's busy bit is set */
  desc = (char *)arrayElementAt(&GDT, task->selector / 8);
  desc += 5;
  temp = *desc;
  if(temp & 2) return 1;
  else return 0;
}
